from gr_nlp_toolkit.data.downloader import Downloader
import os


class DownloaderStub(Downloader):
    def download_processor(self, processor_name: str, target_path: str):
        with open(target_path , 'wb') as f:
            pass


# The labels for the output of the ner model.
# A string label can be obtained by an output index
ner_labels = ['O',
           'S-GPE',
           'S-ORG',
           'S-CARDINAL',
           'B-ORG',
           'E-ORG',
           'B-DATE',
           'E-DATE',
           'S-NORP',
           'B-GPE',
           'E-GPE',
           'S-EVENT',
           'S-DATE',
           'S-PRODUCT',
           'S-LOC',
           'I-ORG',
           'S-PERSON',
           'S-ORDINAL',
           'B-PERSON',
           'I-PERSON',
           'E-PERSON',
           'B-LAW',
           'I-LAW',
           'E-LAW',
           'B-MONEY',
           'I-MONEY',
           'E-MONEY',
           'B-EVENT',
           'I-EVENT',
           'E-EVENT',
           'B-FAC',
           'E-FAC',
           'I-DATE',
           'S-PERCENT',
           'B-QUANTITY',
           'E-QUANTITY',
           'B-WORK_OF_ART',
           'I-WORK_OF_ART',
           'E-WORK_OF_ART',
           'I-FAC',
           'S-LAW',
           'S-TIME',
           'B-LOC',
           'E-LOC',
           'I-LOC',
           'S-FAC',
           'B-TIME',
           'E-TIME',
           'S-WORK_OF_ART',
           'B-PRODUCT',
           'E-PRODUCT',
           'B-CARDINAL',
           'E-CARDINAL',
           'S-MONEY',
           'S-LANGUAGE',
           'I-TIME',
           'I-PRODUCT',
           'I-GPE',
           'I-QUANTITY',
           'B-NORP',
           'E-NORP',
           'S-QUANTITY',
           'B-PERCENT',
           'I-PERCENT',
           'E-PERCENT',
           'I-CARDINAL',
           'B-ORDINAL',
           'I-ORDINAL',
           'E-ORDINAL']



from gr_nlp_toolkit.data.downloader import Downloader
import gdown


class GDriveDownloader(Downloader):
    def __init__(self):
        self.urls = {
            'pos': 'https://drive.google.com/uc?id=1Or5HDk1kVnxI3_w0fwgR8-dzO0jvcc_L',  # pos link
            'ner': 'https://drive.google.com/uc?id=1fx0pHtcN7F2Vj9L8y5TUpbjSqKTUaT3i',  # ner link
            'dp': 'https://drive.google.com/uc?id=1NhEqmLBf67Ydw-LdI7eB-f0afMPgNSmG'  # dp link
        }

    def download_processor(self, processor_name: str, target_path: str):
        gdown.download(self.urls[processor_name], output=target_path, quiet=False)



import string
from collections import Counter
from torch.utils.data import Dataset
import torch
import torch.nn as nn

class TextVectorizer:
    """
    Used to vectorize given text based on a learned vocabulary.
    After training the vocabulary on a corpus, the resulting encoding is:
    1                         : Padding
    [1 to max_vocab_size+1]   : tokens learnt in the vocab. (This could be smaller than the actual max number provided)
    len(vocab)-1              : [SOS] symbol
    len(vocab)                : OOV tokens
    """
    def __init__(self, mode):
        """
        vocab: dictionary (token --> index)
        idx2token: list (idx --> index)
        :param mode:
        """
        assert mode in {"word", "char"}
        self.vocab = dict()
        self.idx2token = []
        self.mode = mode

    def build_vocab(self, corpus, max_size=25000):
        """
        Builds the vocabulary from a corpus of sentences. The words get encoded by
        count of appearances in the data.
        :param corpus: A list of sentences as strings.
        :param max_size: The max size of words that can be encoded, excluding codes for <s> & OOV tokens.
        """
        counts = Counter()
        self.vocab["<pad>"] = 0
        self.idx2token.append("<pad>")
        idx = 1
        if self.mode == "word":
            # In the case of words, we remove punctuation and split on whitespaces
            for line in corpus:
                # Remove punctuation
                line = line.translate(str.maketrans("", "", string.punctuation))
                # Split the line in whitespaces to get the words
                tokens = line.split()
                # Update counts
                counts.update(tokens)
        # mode == "char"
        else:
            # Here we do not do any regularization, and split on every character.
            for line in corpus:
                tokens = [char for char in line]
                counts.update(tokens)
        # Add the most frequent tokens to the vocabulary.
        for (name, count) in counts.most_common(max_size):
            self.vocab[name] = idx
            self.idx2token.append(name)
            idx += 1
        # Add [SOS] token.
        self.vocab["<s>"] = idx
        self.idx2token.append("<s>")


    def encode_dataset(self, corpus):
        """
        Takes as input a corpus of sentences, generates source/target training pairs
        and encodes them based on the vocabulary. Then it returns the pairs as tuples of tensors.
        :param corpus: Array of sentences in the form of strings.
        :return: list of pairs of torch.LongTensor objects
        """
        # We start by tokenizing the corpus.
        tokenized_dataset = []
        for line in corpus:
            if self.mode == "word":
                # Strip punctuation and split on whitespaces.
                tokens = line.translate(str.maketrans("", "", string.punctuation)).split()
            else:
                # No regularization applied for characters.
                tokens = [char for char in line]
            # Also find the length of the longest sequence, taking into account the addition of a [SOS]/[EOS] symbol.
            tokenized_dataset.append(tokens)
        # Make source & target sentences and encode them based on the dictionary.
        source_vecs, target_vecs = [], []
        for sequence in tokenized_dataset:
            # Ignore strings that may be reduced to empty after stripping punctuation & whitespaces
            # (only happens if mode == "word")
            if not sequence:
                continue
            # Initialize source vectorized sentence with <s> token.
            source_vector = [self.vocab["<s>"]]
            target_vector = []
            for idx in range(len(sequence)-1):
                source_vector.append(self.vocab.get(sequence[idx], len(self.vocab)))
                target_vector.append(self.vocab.get(sequence[idx], len(self.vocab)))
            target_vector.append(self.vocab.get(sequence[-1], len(self.vocab)))
            # Add to sources/targets.
            source_vecs.append(source_vector)
            target_vecs.append(target_vector)

        """# Get the length for each sequence in the data
        source_lengths = torch.LongTensor(list(map(len, source_vecs)))
        target_lengths = torch.LongTensor(list(map(len, target_vecs)))"""
        # Convert data to LongTensors.
        for i in range(len(source_vecs)):
            source_vecs[i] = torch.LongTensor(source_vecs[i])
            target_vecs[i] = torch.LongTensor(target_vecs[i])
        # Pad & Sort sequences
        source_tensors = nn.utils.rnn.pad_sequence(source_vecs, batch_first=True)
        target_tensors = nn.utils.rnn.pad_sequence(target_vecs, batch_first=True)
        # Create Dataset object
        dataset = GreekDataset(source_tensors, target_tensors)
        # Return the Dataset & the sequence lengths (to be used for packing)
        return dataset

    def split_sequence(self, sequence):
        """
        Splits a sequence based on the tokenization mode configured, and returns it without indexing it.
        """
        if self.mode == "word":
            tokens = sequence.translate(str.maketrans("", "", string.punctuation)).split()
        else:
            tokens = [char for char in sequence]

        return sequence

    def input_tensor(self, sequence):
        """
        Takes a sentence and returns its encoding, based on the vocabulary, to be used for inference.
        :param sequence: (String) The sentence to be encoded.
        :return: Encoded sentence in form of a torch.Longtensor object.
        """
        if self.mode == "word":
            tokens = sequence.translate(str.maketrans("", "", string.punctuation)).split()
        else:
            tokens = [char for char in sequence]
        vectorized_input = []
        for token in tokens:
            vectorized_input.append(self.vocab.get(token, len(self.vocab)))

        # Convert to tensor
        vectorized_input = torch.LongTensor(vectorized_input)

        return vectorized_input


class GreekDataset(Dataset):

    def __init__(self, source_vecs, target_vecs):
        """
        Gets two arrays of source and target vectors and outputs a Dataset object of those arrays
        :param source_vecs: array of source vectors
        :param target_vecs: array of target vectors
        """
        self.n_samples = source_vecs.size(0)
        self.x_data = source_vecs
        self.y_data = target_vecs


    def __getitem__(self, index):
        return self.x_data[index], self.y_data[index]

    def __len__(self):
        return self.n_samples


from gr_nlp_toolkit.processors.abstract_processor import AbstractProcessor
from gr_nlp_toolkit.domain.document import Document
from gr_nlp_toolkit.models import g2g_RBNLM_model
from gr_nlp_toolkit.domain.textVectorizer import TextVectorizer
from gr_nlp_toolkit.models.g2g_RBNLM_model import LanguageModel
from gr_nlp_toolkit.models.g2g_transformer_model import ByT5Model
import torch
import pickle

def detect_language(text):
    """
    Checks whether the majority of the letters in the input text are in the greek or the latin script
    It is used to identify whether the text is in greek or greeklish (latin script), in order to skip unnecessary conversions.

    Args:
        text (str): The input text

    Returns:
        script (str): The dominant script
    """
    # Filter out non-letter characters
    valid_characters = [char for char in text if char.isalpha()]
    
    # Count Greek and English letters
    greek_count = sum(1 for char in valid_characters if '\u0370' <= char <= '\u03FF' or '\u1F00' <= char <= '\u1FFF')
    english_count = sum(1 for char in valid_characters if '\u0041' <= char <= '\u005A' or '\u0061' <= char <= '\u007A')
    
    script = "greek" if greek_count >= english_count else "latin"
    return script


class G2G(AbstractProcessor):
    """
        Greeklsih to Greek (G2G) processor class.

        This class performs G2G conversion using either an LSTM-based model or a transformer-based model.
        It initializes the model, loads the necessary components, and provides functionality to process documents
        and convert text using the specified mode.
    """

    def __init__(self, mode = 'LSTM', model_path = None, tokenizer_path = None, device = 'cpu'):

        """
        Initializes the G2G class with the specified parameters.

        Args:
            mode (str, optional): The mode of the model, either 'LSTM' or 'transformer'. Defaults to 'LSTM'.
            model_path (str, optional): Path to the pre-trained model. Defaults to None.
            tokenizer_path (str, optional): Path to the tokenizer for LSTM mode. Defaults to None.
            device (str, optional): Device to load the model on ('cpu' or 'cuda'). Defaults to 'cpu'.
        """

        
        self.mode = mode
        self.device = torch.device(device)
        
        if self.mode == 'LSTM':
            # Define the model parameters (more info: https://aclanthology.org/2024.lrec-main.1330/)
            input_size = 120
            embed_size = 32
            hidden_size = 512
            output_size = 120
            
            # Load and initialize the LSTM model
            self.beam_size = 5
            self.model = g2g_RBNLM_model.LSTM_LangModel(input_size, embed_size, hidden_size, output_size)
            

            # Load and initialize the tokenizer
            self.text_vectorizer = TextVectorizer("char")
            
            if(model_path is not None):
                self.model.load_state_dict(torch.load(model_path, map_location=self.device, weights_only=True))

            
            if(tokenizer_path is not None):
                with open(tokenizer_path, "rb") as file:
                    self.text_vectorizer = pickle.load(file)

            # Initialize the LanguageModel
            self.LM = LanguageModel(self.text_vectorizer, self.model, device=self.device)
                    

        elif self.mode == 'transformer':
            self.model = ByT5Model(model_path, device=self.device)
            self.model.eval()


    def __call__(self, doc: Document) -> Document:
        """
        Processes a document to perform Greeklish to Greek conversion.

        Args:
            doc (Document): The document to process.

        Returns:
            Document: The document with text converted using the specified model.
        """

        # If the text is in already in greek, skip the g2g conversion
        if(detect_language(doc.text) == 'greek'):
            return doc
        
        

        # Perform G2G conversion based on the mode
        if(self.mode == 'LSTM'):
            doc.text = self.LM.translate([doc.text], self.beam_size)[0]
        elif(self.mode == 'transformer'):
            doc.text = self.model(doc.text)

        return doc


from torch import nn
from gr_nlp_toolkit.models.util import create_mask_from_length

class POSModel(nn.Module):
    """
    Part-Of-Speech (POS) tagging model class based on BERT.

    This class defines a POS model using a pre-trained BERT model 
    with a dropout and multiple linear layers on top.

    Attributes:
        _bert_model (AutoModel): The pre-trained BERT model.
        _dp (nn.Dropout): Dropout layer for regularization.
        _linear_dict (nn.ModuleDict): Dictionary of linear layers for different features.
    """

    def __init__(self, bert_model, feat_to_size, dp):
        """
        Initializes the POSModel with the specified parameters.

        Args:
            bert_model (AutoModel): The pre-trained BERT model.
            feat_to_size (dict): A dictionary mapping feature names to the size of their output layers.
            dp (float): Dropout probability.
        """

        super(POSModel, self).__init__()
        self._bert_model = bert_model
        self._dp = nn.Dropout(dp)

        self._linear_dict = nn.ModuleDict({feat: nn.Linear(768, feat_to_size[feat]) for feat in feat_to_size})

    def forward(self, text, text_len):
        """
        Performs a forward pass of the model.

        Args:
            text (torch.Tensor): Input tensor containing token IDs.
            text_len (torch.Tensor): Tensor containing the lengths of each sequence in the batch.

        Returns:
            dict: A dictionary containing the output tensors for each feature.
        """
        
        attention_mask = create_mask_from_length(text_len, text.shape[1])
        bert_output = self._dp(self._bert_model(text, attention_mask=attention_mask)[0])

        output_dict = {feat: self._linear_dict[feat](bert_output) for feat in self._linear_dict}
        return output_dict




from transformers import T5ForConditionalGeneration, AutoTokenizer
from torch import nn
import torch



class ByT5Model(nn.Module):
    """
    A wrapper class for the T5 model for conditional generation

    Attributes:
        model (T5ForConditionalGeneration): The pre-trained ByT5 model
        tokenizer (AutoTokenizer): The tokenizer for the ByT5 model
    """

    def __init__(self, model_path = None, device = 'cpu'):
        """
        Initializes the ByT5Model with a pretrained T5 model and tokenizer.

        Args:
            model_path: The path to the pretrained model and tokenizer.
        """
        super(ByT5Model, self).__init__()
        
        self.model = T5ForConditionalGeneration.from_pretrained(model_path)
        self.tokenizer = AutoTokenizer.from_pretrained(model_path)
        self.device = torch.device(device)
        self.model.to(self.device)

    def forward(self, text):
        """
        Performs inference to the ByT5 to generate the transliterated text

        Args:
            text: the input text in greeklish

        Returns:
            The output text in greek
        """
        self.model.eval()
        tokenized_text = self.tokenizer(text, return_tensors="pt").input_ids

        output = self.model.generate(tokenized_text.to(self.device), max_length=10000)

        return self.tokenizer.decode(output[0], skip_special_tokens=True)



from gr_nlp_toolkit.domain.document import Document
from gr_nlp_toolkit.processors.dp import DP
from gr_nlp_toolkit.processors.ner import NER
from gr_nlp_toolkit.processors.pos import POS
from gr_nlp_toolkit.processors.g2g import G2G

from gr_nlp_toolkit.processors.tokenizer import Tokenizer
from huggingface_hub import hf_hub_download

from typing import Literal
import torch

from transformers import logging
logging.set_verbosity_error()

def get_device_name() -> Literal["mps", "cuda", "cpu"]:
    """
    Returns the name of the device where this module is running.

    This is a simple implementation that doesn't cover cases when more powerful GPUs are available 
    and not a primary device ('cuda:0') or MPS device is available but not configured properly:
    https://pytorch.org/docs/master/notes/mps.html

    Returns:
        Literal["mps", "cuda", "cpu"]: Device name, like 'cuda' or 'cpu'.

    Examples:
        >>> torch.cuda.is_available = lambda: True
        >>> torch.backends.mps.is_available = lambda: False
        >>> get_device_name()
        'cuda'

        >>> torch.cuda.is_available = lambda: False
        >>> torch.backends.mps.is_available = lambda: True
        >>> get_device_name()
        'mps'

        >>> torch.cuda.is_available = lambda: False
        >>> torch.backends.mps.is_available = lambda: False
        >>> get_device_name()
        'cpu'
    """
    if torch.cuda.is_available():
        return "cuda"
    elif torch.backends.mps.is_available():
        return "mps"
    else:
        return "cpu"


class Pipeline:
    """
    The central class of the toolkit. A pipeline is created after a list of processors are specified. The user can
    then annotate a document by using the __call__ method of the Pipeline
    
    Attributes:
        _processors: A list of the processors that will be used in the pipeline
        _processor_cache: A ProcessorCache object that is used to download the processors
        device: The device where the pipeline will run

    """

    def __init__(self, processors: str, use_cpu: bool = False):
        """ 
        Initializes the pipeline with the specified processors

        Args:
            processors: A string with the names of the processors you want to load, available values: 'ner', 'por', 'dp
            use_cpu: A boolean that specifies if the pipeline will run on the CPU
        """

        # if the user wants to use the CPU, we set the device to 'cpu'
        if(use_cpu):
            self.device = "cpu"
        else:
            self.device = get_device_name()

        self._processors = []

        processors = set(processors.split(","))

        # ner: Named Entity Recognition Processor 
        # pos: Part of Speech Recognition Processor
        # dp: Dependency Parsing 
        # g2g: Greeklish to Greek Transliteration Processor (ByT5 model)
        # g2g_lite: Greeklish to Greek Transliteration Processor (LSTM model)
        available_processors = ['ner', 'pos', 'dp', 'g2g_lite', 'g2g']


        # Adding the g2g processor, which must be the first in the pipeline
        if("g2g_lite" in processors):
            self._processors.append(G2G(mode="LSTM", model_path="gr_nlp_toolkit/RBNLM_weights/LSTM_LM_50000_char_120_32_512.pt", tokenizer_path="gr_nlp_toolkit/RBNLM_weights/RBNLMtextVectorizer.pkl", device=self.device))
            processors.remove("g2g_lite")
        elif("g2g" in processors):
            self._processors.append(G2G(mode="transformer", model_path="AUEB-NLP/ByT5_g2g", device=self.device))
            processors.remove("g2g")

            
        # Adding the tokenizer processor
        self._processors.append(Tokenizer())
        for p in processors:
            if p == available_processors[0]:
                ner_path = hf_hub_download(repo_id="AUEB-NLP/gr-nlp-toolkit", filename="ner_processor")
                self._processors.append(NER(model_path=ner_path, device=self.device))
            elif p == available_processors[1]:
                pos_path = hf_hub_download(repo_id="AUEB-NLP/gr-nlp-toolkit", filename="pos_processor")
                self._processors.append(POS(model_path=pos_path, device=self.device))
            elif p == available_processors[2]:
                dp_path = hf_hub_download(repo_id="AUEB-NLP/gr-nlp-toolkit", filename="dp_processor")
                self._processors.append(DP(model_path=dp_path, device=self.device))
            else:
                raise Exception(f"Invalid processor name, please choose one of {available_processors}")

    def __call__(self, text: str) -> Document:
        
        """
        Annotate a text with the processors present in the pipeline

        Args:
            text: The text that will be annotated
        """

        # Create a document from the text
        self._doc = Document(text)

        # Pass the document through every processor
        for processor in self._processors:
            # print(processor)
            processor(self._doc)

        return self._doc
    
if __name__ == "__main__": 


    nlp = Pipeline("g2g,ner,dp,pos")
   
    txts = ["Uparxoun autoi pou kerdizoun apo mia katastash kai autoi pou hanoun",
            "o volos kai h larisa einai poleis ths thessalias",
            "Η Αθήνα είναι η μεγαλύτερη πόλη της Ελλάδας"]

    for txt in txts:

        doc = nlp(txt)
        
        print(doc.text)
        for token in doc.tokens:
            print(f"{token.text}: {token.ner}, {token.upos}, {token.feats}, {token.head}, {token.deprel}") # the text of the token
            




import unittest

from gr_nlp_toolkit.processors.tokenizer import *


class TestTokenizer(unittest.TestCase):
    def test_strip_accents_and_lowercase1(self):
        result = strip_accents_and_lowercase('ποιητής')
        self.assertEqual('ποιητης', result)

    def test_strip_accents_and_lowercase2(self):
        result = strip_accents_and_lowercase('ΠΟΙΗΤΗΣ')
        self.assertEqual('ποιητης', result)

    """"
        Tests with no sub-words:
    """

    def test_create_ids_without_subwords(self):
        ids = create_ids('ο ποιητης')
        # 2 special tokens + 2 given words
        self.assertEqual(4, len(ids))

    def test_create_tokens_without_subwords(self):
        ids = [101, 247, 6981, 102]
        tokens = convert_to_tokens(ids)
        # 2 words, special tokens are not included
        self.assertEqual(2, len(tokens))
        self.assertEqual('ο', tokens[0])
        self.assertEqual('ποιητης', tokens[1])

    def test_create_mask_and_tokens_without_subwords(self):
        tokens = ['ο', 'ποιητης']
        mask, tokens, subword2word = create_mask_and_tokens(tokens, [247, 6981])

        self.assertEqual(2, len(mask))
        self.assertEqual([True, True], mask)
        self.assertEqual(2, len(tokens))
        self.assertEqual(1, len(tokens[0].subwords))
        self.assertEqual(1, len(tokens[1].subwords))
        self.assertEqual(247, tokens[0]._ids[0])
        self.assertEqual(6981, tokens[1]._ids[0])
        self.assertEqual('ο', tokens[0].subwords[0])
        self.assertEqual('ποιητης', tokens[1].subwords[0])
        self.assertEqual('ο', tokens[0].text)
        self.assertEqual('ποιητης', tokens[1].text)
        self.assertEqual(len(subword2word.keys()), 3)
        self.assertEqual(subword2word[1], 1)
        self.assertEqual(subword2word[2], 2)

    """"
        Tests with sub-words:
    """

    def test_create_ids_with_subwords(self):
        ids = create_ids('ενα ποιηματακι')
        # 2 special tokens + 1 word without sub-words + 1 word with 1 sub-word
        self.assertEqual(5, len(ids))

    def test_create_tokens_with_subwords(self):
        ids = [101, 370, 6623, 701, 102]
        tokens = convert_to_tokens(ids)
        # 2 words, special tokens are not included
        self.assertEqual(3, len(tokens))
        self.assertEqual('ενα', tokens[0])
        self.assertEqual('ποιηματα', tokens[1])
        self.assertEqual('##κι', tokens[2])

    def test_create_mask_and_tokens_with_subwords(self):
        tokens = ['ενα', 'ποιηματα', '##κι']
        mask, tokens, subword2word = create_mask_and_tokens(tokens, [370, 6623, 701])

        self.assertEqual(3, len(mask))
        self.assertEqual([True, True, False], mask)
        self.assertEqual(2, len(tokens))
        self.assertEqual(1, len(tokens[0].subwords))
        self.assertEqual(2, len(tokens[1].subwords))
        self.assertEqual(370, tokens[0]._ids[0])
        self.assertEqual(6623, tokens[1]._ids[0])
        self.assertEqual(701, tokens[1]._ids[1])
        self.assertEqual('ενα', tokens[0].subwords[0])
        self.assertEqual('ποιηματα', tokens[1].subwords[0])
        self.assertEqual('##κι', tokens[1].subwords[1])
        self.assertEqual('ενα', tokens[0].text)
        self.assertEqual('ποιηματακι', tokens[1].text)
        self.assertEqual(len(subword2word.keys()), 4)
        self.assertEqual(subword2word[1], 1)
        self.assertEqual(subword2word[2], 2)
        self.assertEqual(subword2word[3], 2)

    def test_tokenizer(self):
        tokenizer = Tokenizer()
        doc = tokenizer(Document('Ο ποιητής'))
        # document has all field set
        self.assertIsNotNone(doc.text)
        self.assertIsNotNone(doc.input_ids)
        self.assertIsNotNone(doc.token_mask)
        self.assertIsNotNone(doc.tokens)
        self.assertIsNotNone(doc.subword2word)

    def test_create_dataset_and_dataloader(self):
        input_ids = [101, 370, 6623, 701, 102]
        dataset, dataloader = create_dataset_and_dataloader(input_ids)
        self.assertIsNotNone(dataset.input_ids)
        self.assertIsNotNone(dataloader.dataset)
        self.assertEqual(dataset, dataloader.dataset)
        self.assertEqual(dataset.input_ids, [input_ids])


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import shutil

from gr_nlp_toolkit.data.downloader_stub import DownloaderStub
from gr_nlp_toolkit.data.processor_cache import ProcessorCache


class TestProcessorCache(unittest.TestCase):
    def test_download_processors_sequentially(self):

        sep = os.sep
        cache_path = "./test"
        stub = DownloaderStub()
        processor_cache = ProcessorCache(stub , cache_path)
        processor_cache.get_processor_path('ner')
        self.assertTrue(os.path.exists(cache_path + sep + "ner_processor"))
        processor_cache.get_processor_path('pos')
        self.assertTrue(os.path.exists(cache_path + sep + "pos_processor"))
        dp_path = processor_cache.get_processor_path('dp')
        self.assertTrue(type(dp_path) == str)
        self.assertTrue(os.path.exists(cache_path + sep + "dp_processor"))
        self.assertTrue(dp_path == (cache_path + sep + "dp_processor"))
        # Remove any files created
        shutil.rmtree(cache_path)

    def test_download_processor_removing_file_and_folder(self):

        home = os.path.expanduser("~")
        sep = os.sep
        cache_path = "./test"
        stub = DownloaderStub()
        processor_cache = ProcessorCache(stub, cache_path)
        processor_cache.get_processor_path('ner')
        self.assertTrue(os.path.exists(cache_path + sep + "ner_processor"))
        os.remove(cache_path + sep + "ner_processor")
        # Assert that the file is removed
        self.assertTrue(not os.path.exists(cache_path + sep + "ner_processor"))
        processor_cache.get_processor_path('ner')
        # Assert that the file has appeared again
        self.assertTrue(os.path.exists(cache_path + sep + "ner_processor"))
        processor_cache.get_processor_path('pos')
        # Remove entire directory
        shutil.rmtree(cache_path)
        processor_cache.get_processor_path('pos')
        # Assert that the certain processor has appeared again
        self.assertTrue(os.path.exists(cache_path + sep + "pos_processor"))
        # Remove any files created
        shutil.rmtree(cache_path)

if __name__ == '__main__':
    unittest.main()



from typing import List


class Token:
    """
    Token class which represents a word/token
    """

    def __init__(self, subwords: List[str]):
        """
            Create a Token object setting possible parameters other than the text as None

            Keyword arguments:
            subwords: A list with the token's subwords
        """

        # the text
        self._text = ""

        # the subwords
        self._subwords = subwords

        # the ids
        self._ids = []

        # Named Entity Recognition parameters
        # the named entity
        self._ner = None

        # Part of Speech Tagging parameters
        # the universal pos tag
        self._upos = None
        # the universal morphological features
        self._feats = {}

        # Dependency Parsing parameters
        # the dependant word index in the sentence
        self._head = None
        # the label of the relation between the specific word and the dependant one
        self._deprel = None

    @property
    def text(self):
        """
        The text
        """
        return self._text

    @text.setter
    def text(self, value):
        self._text = value

    @property
    def subwords(self):
        """
        A list with Token's subwords
        """
        return self._subwords

    @subwords.setter
    def subwords(self, value):
        self._subwords = value

    @property
    def ids(self):
        return self._ids

    @ids.setter
    def ids(self, value):
        self._ids = value

    @property
    def ner(self):
        """
        The Named Entity
        """
        return self._ner

    @ner.setter
    def ner(self, value):
        self._ner = value

    @property
    def upos(self):
        """
        The universal pos tag
        """
        return self._upos

    @upos.setter
    def upos(self, value):
        self._upos = value

    @property
    def feats(self):
        """
        The universal morphological features
        """
        return self._feats

    @feats.setter
    def feats(self, value):
        self._feats = value

    @property
    def head(self):
        """
        The dependant word index in the sentence
        """
        return self._head

    @head.setter
    def head(self, value):
        self._head = value

    @property
    def deprel(self):
        """
        The label of the relation between the specific word and the dependant one
        """
        return self._deprel

    @deprel.setter
    def deprel(self, value):
        self._deprel = value



import torch
from torch import nn

from transformers import AutoModel

from gr_nlp_toolkit.configs.ner_labels import ner_labels
from gr_nlp_toolkit.domain.document import Document
from gr_nlp_toolkit.processors.abstract_processor import AbstractProcessor

from gr_nlp_toolkit.models.ner_model import NERBERTModel


model_params = {'dp': 0}


class NER(AbstractProcessor):
    """
    Named Entity Recognition (NER) processor class.

    This class performs NER using a pre-trained BERT model. It initializes the model,
    loads the necessary components, and provides functionality to process documents
    and perform NER on them.

    Attributes:
        I2L (list): A list of label names for the NER task.
        output_size (int): The number of output labels.
        _model (NERBERTModel): The NER model based on BERT.
        softmax (nn.Softmax): Softmax function for output normalization.
        device (torch.device): Device on which the model is loaded.
    """

    def __init__(self, model_path=None, device='cpu', entities=18,):
        """
        Initializes the NER class with the specified parameters.

        Args:
            model_path (str, optional): Path to the pre-trained model. Defaults to None.
            device (str, optional): Device to load the model on ('cpu' or 'cuda'). Defaults to 'cpu'.
            entities (int, optional): Number of entity labels. Should be set to 18. Defaults to 18.

        Raises:
            ValueError: If the number of entities is not 18.
        """

        # Entities are the semantic catgories of the NER task (more info: http://nlp.cs.aueb.gr/theses/smyrnioudis_bsc_thesis.pdf)
        if entities == 18:
            self.I2L = ner_labels
            self.output_size = len(self.I2L)
        else:
            raise ValueError('Entities should be set to 18')

         # Initialize the BERT model
        bert_model = AutoModel.from_pretrained('nlpaueb/bert-base-greek-uncased-v1')
        self._model = NERBERTModel(bert_model, self.output_size, **model_params)
        self.softmax = nn.Softmax(dim=-1)
        self.device = torch.device(device)
        self._model.to(self.device)
        self._model.eval()

        # load the pretrained model if provided
        if model_path is not None:
            self._model.load_state_dict(torch.load(model_path, map_location=self.device, weights_only=True), strict=False)



    def __call__(self, doc: Document) -> Document:
        """
        Processes a document to perform Named Entity Recognition.

        Args:
            doc (Document): The document to process.

        Returns:
            Document: The document with NER tags assigned to the tokens.
        """
         
        # Get the input ids and text length of the document
        input_ids, text_len = next(iter(doc.dataloader))['input']
        
        # Perform NER with the model
        output = self._model(input_ids.to(self.device), text_len.to(self.device))
        predictions = self.softmax(output)
        predictions = torch.argmax(predictions[0], axis=-1).detach().cpu().numpy()

        # map predictions -> tokens, special tokens are not included
        i = 0
        for mask, pred in zip(doc.token_mask, predictions[1: len(predictions) - 1]):
            if mask:
                token = doc.tokens[i]
                token.ner = self.I2L[pred]
                i+=1
        
        return doc


import os
from pathlib import Path
from os.path import expanduser

from gr_nlp_toolkit.data.downloader import Downloader


class ProcessorCache:
    def __init__(self, downloader : Downloader, cache_path : str):
        """
        Initializes the cache of processors creating necessary directories
        :param downloader: an object with the Downloader interface
        """
        # Get home directory
        self.home = expanduser("~")
        self.sep = os.sep
        self.cache_path = cache_path
        self.downloader = downloader
        # Initialize the filenames for each processor
        self.processor_names_to_filenames = {
            'ner': 'ner_processor',
            'pos': 'pos_processor',
            'dp': 'dp_processor'
        }
        self.update_cache_path()

    def update_cache_path(self):
        Path(self.cache_path).mkdir(parents=True, exist_ok=True)

    def get_processor_path(self, processor_name: str) -> str:
        # Update cache path in case any changes occured
        self.update_cache_path()
        target_filename = self.processor_names_to_filenames[processor_name]
        if not os.path.exists(self.cache_path + self.sep + target_filename):
            self.downloader.download_processor(processor_name, self.cache_path + self.sep + target_filename)
        # Return the path
        return self.cache_path + self.sep + target_filename



import torch
from torch import nn

from transformers import AutoModel

from gr_nlp_toolkit.configs.dp_labels import dp_labels
from gr_nlp_toolkit.domain.document import Document
from gr_nlp_toolkit.processors.abstract_processor import AbstractProcessor

from gr_nlp_toolkit.models.dp_model import DPModel


class DP(AbstractProcessor):
    """
    Dependency Parsing (DP) processor class.

    This class performs dependency parsing using a pre-trained BERT model. It initializes the model,
    loads the necessary components, and provides functionality to process documents
    and assign head and dependency relation (deprel) tags to tokens.

     Attributes:
        I2L (list): A list of dependency relation labels.
        output_size (int): The number of output labels.
        _model (DPModel): The dependency parsing model based on BERT.
        softmax (nn.Softmax): Softmax function for output normalization.
        device (torch.device): Device on which the model is loaded.
    """

    def __init__(self, model_path=None, device='cpu'):
        """
        Initializes the DP class with the specified parameters.

        Args:
            model_path (str, optional): Path to the pre-trained model. Defaults to None.
            device (str, optional): Device to load the model on ('cpu' or 'cuda'). Defaults to 'cpu'.
        """

        self.I2L = dp_labels
        self.output_size = len(self.I2L)

         # Initialize the BERT model
        bert_model = AutoModel.from_pretrained('nlpaueb/bert-base-greek-uncased-v1')
        self._model = DPModel(bert_model, self.I2L, 0)

        self.softmax = nn.Softmax(dim=-1)
        self.device = torch.device(device)
        self._model.to(self.device)
        self._model.eval()

        # Load the pretrained model if provided
        if model_path is not None:
            self._model.load_state_dict(torch.load(model_path, map_location=self.device, weights_only=True), strict=False)

    def __call__(self, doc: Document) -> Document:
        """
        Processes a document to perform dependency parsing.

        Args:
            doc (Document): The document to process.

        Returns:
            Document: The document with head and deprel tags assigned to the tokens.
        """

        # Predict heads
        input_ids, text_len = next(iter(doc.dataloader))['input']

        output_heads = 'heads'
       
        predictions_heads = self._model(input_ids.to(self.device), text_len.to(self.device))
        predictions_heads = self.softmax(predictions_heads[output_heads])
        predictions_heads = torch.argmax(predictions_heads[0], axis=-1).detach().cpu().numpy()

        # Predict dependency relations (deprels)
        output_deprels = 'gathered_deprels'
        
        predictions_deprels = self._model(input_ids.to(self.device), text_len.to(self.device))
        predictions_deprels = self.softmax(predictions_deprels[output_deprels])
        predictions_deprels = torch.argmax(predictions_deprels[0], axis=-1).detach().cpu().numpy()

        # map predictions -> tokens, special tokens are not included
        i = 0
        for mask, pred_head, pred_deprel in zip(doc.token_mask, predictions_heads[1: len(predictions_heads) - 1],
                                                       predictions_deprels[1: len(predictions_deprels) - 1]):
            if mask:
                token = doc.tokens[i]
                token.head = doc.subword2word[pred_head]
                token.deprel = self.I2L[pred_deprel]
                i +=1

        return doc



# import pytorch_wrapper.functional as pwF

from torch import nn
from gr_nlp_toolkit.models.util import create_mask_from_length


class NERBERTModel(nn.Module):
    """
    Named Entity Recognition (NER) model class based on BERT.

    This class defines a NER model using a pre-trained BERT model 
    with a dropout and a linear layer on top.

    Attributes:
        _bert_model (AutoModel): The pre-trained BERT model.
        _dp (nn.Dropout): Dropout layer for regularization.
        _output_linear (nn.Linear): Linear layer to produce model outputs.
    """
    def __init__(self, bert_model, model_output_size, dp):
        """
        Initializes the NERBERTModel with the specified parameters.

        Args:
            bert_model (AutoModel): The pre-trained BERT model.
            model_output_size (int): The size of the output layer.
            dp (float): Dropout probability.
        """
         
        super(NERBERTModel, self).__init__()
        self._bert_model = bert_model
        self._dp = nn.Dropout(dp)
        self._output_linear = nn.Linear(768, model_output_size)

    def forward(self, text, text_len):
        """
        Performs a forward pass of the model.

        Args:
            text (torch.Tensor): Input tensor containing token IDs.
            text_len (torch.Tensor): Tensor containing the lengths of each sequence in the batch.

        Returns:
            torch.Tensor: The output of the linear layer after applying dropout and BERT.
        """

        # Create attention mask
        attention_mask = create_mask_from_length(text_len, text.shape[1])

        return self._output_linear(self._dp(self._bert_model(text, attention_mask=attention_mask)[0]))
    



import torch
from torch.utils.data import Dataset
from torch.utils.data.dataset import T_co


class DatasetImpl(Dataset):
    def __init__(self, input_ids):
        self._input_ids = input_ids

    def __getitem__(self, index) -> T_co:
        return {
            "input": [torch.tensor(self._input_ids[index], dtype=torch.long), torch.tensor(len(self._input_ids[index]))]
        }

    def __len__(self):
        return 1

    @property
    def input_ids(self):
        return self._input_ids

    @input_ids.setter
    def input_ids(self, value):
        self._input_ids = value





class Document:
    """
    Document class that represents an annotated text
    """

    def __init__(self, text: str):
        """
        Create a Document object setting possible parameters other than the text as None

        Keyword arguments:
        param text: The text of the document
        """
        self._text = text

        self._input_ids = None
        self._token_mask = None

        self._tokens = None

        self._dataloader = None

        self._subword2word = None


    @property
    def text(self):
        """
        Return the original text of the document
        """
        return self._text

    @text.setter
    def text(self, value):
        self._text = value


    @property
    def tokens(self):
        """
        A list of Tokens containing the tokens of the text as well as token level annotations
        """
        return self._tokens

    @tokens.setter
    def tokens(self, value):
        self._tokens = value


    @property
    def input_ids(self):
        """
         A tensor of shape [1,mseq] containing the input ids created with the BERT tokenizer
        """
        return self._input_ids

    @input_ids.setter
    def input_ids(self, value):
        self._input_ids = value


    @property
    def token_mask(self):
        """
        A tensor of shape [1,mseq] containign zeros at the positions of the input_ids tensor that map to subword tokens that are non first subword tokens
        """
        return self._token_mask

    @token_mask.setter
    def token_mask(self, value):
        self._token_mask = value


    @property
    def dataloader(self):
        return self._dataloader

    
    @dataloader.setter
    def dataloader(self, value):
        self._dataloader = value

    @property
    def subword2word(self):
        """
        A mapping for each subword to the word
        """
        return self._subword2word

    @subword2word.setter
    def subword2word(self, value):
        self._subword2word = value


from typing import List, Tuple, Dict
from torch.utils.data import DataLoader, Dataset

from gr_nlp_toolkit.domain.dataset import DatasetImpl
from gr_nlp_toolkit.domain.document import Document
from gr_nlp_toolkit.processors.abstract_processor import AbstractProcessor

import unicodedata

from gr_nlp_toolkit.domain.token import Token

from transformers import AutoTokenizer

tokenizer_greek = AutoTokenizer.from_pretrained('nlpaueb/bert-base-greek-uncased-v1')


def strip_accents_and_lowercase(s: str) -> str:
    """
    Strips accents from a string and converts it to lowercase.

    Args:
        s: A string from which to strip accents.

    Returns:
        A new string with accents removed and converted to lowercase.
    """
    return ''.join(c for c in unicodedata.normalize('NFD', s)
                   if unicodedata.category(c) != 'Mn').lower()


def create_ids(text: str) -> List[int]:
    """
    Encodes a given text into a list of token IDs using a tokenizer.

    Args:
        text: A string to encode.

    Returns:
        A list of token IDs.
    """
    return tokenizer_greek.encode(text)


def create_text(ids: List[int]) -> List[int]:
    """
    Decodes a list of token IDs back into a text string.

    Args:
        ids: A list of token IDs to decode.

    Returns:
        A decoded string with special tokens skipped.
    """
    return tokenizer_greek.decode(ids, skip_special_tokens=True)


def convert_to_tokens(input_ids: List[int]) -> List[str]:
    """
    Converts a list of token IDs into their corresponding token strings.

    Args:
        input_ids: A list of token IDs.

    Returns:
        A list of token strings with special tokens skipped.
    """
    return tokenizer_greek.convert_ids_to_tokens(input_ids, skip_special_tokens=True)


def remove_special_tokens(input_ids: List[int]) -> List[int]:
    """
    Removes special tokens from a list of token IDs.

    Args:
        input_ids: A list of token IDs.

    Returns:
        A new list of token IDs with special tokens removed.
    """
    input_ids_without_special_tokens = []
    for input_id in input_ids:
        if input_id not in tokenizer_greek.all_special_ids:
            input_ids_without_special_tokens.append(input_id)
    return input_ids_without_special_tokens



def create_mask_and_tokens(input_tokens: List[str], input_ids: List[int]) -> Tuple[List[str], List[Token], Dict]:
    """
    Creates a mask, tokens, and subword-to-word mapping from input tokens and IDs.

    Args:
        input_tokens: A list of input token strings.
        input_ids: A list of input token IDs.

    Returns:
        A tuple containing:
        - A list of booleans indicating whether each token is a subword.
        - A list of Token objects.
        - A dictionary mapping subword indices to word indices.
    """
    mask = []
    tokens = []
    subword2word = {}

    word = 0
    # for each token
    for j, input in enumerate(zip(input_tokens, input_ids), 1):
        t = input[0]
        i = input[1]
        # it isn't a sub-word
        if not t.startswith("##"):
            # create a token object
            tokenObj = Token([t])
            tokenObj.ids.append(i)
            tokens.append(tokenObj)
            mask.append(True)
            word = word + 1
        else:

            # add sub-words to token
            tokenObj.subwords.append(t)
            tokenObj.ids.append(i)
            mask.append(False)
        subword2word[j] = word

    # create text
    for token in tokens:
        token.text = create_text(token.ids)

    # Adding a 0-0 mapping to subword2word
    subword2word[0] = 0

    return mask, tokens, subword2word


def create_dataset_and_dataloader(input_ids) -> Tuple[Dataset, DataLoader]:
    """
    Creates a dataset and dataloader from input IDs.

    Args:
        input_ids: A list of input token IDs.

    Returns:
        A tuple containing:
        - A Dataset object.
        - A DataLoader object.
    """
    dataset = DatasetImpl([input_ids])
    dataloader = DataLoader(dataset)
    return dataset, dataloader



class Tokenizer(AbstractProcessor):
    """
    Tokenizer class that takes a document as an input with the text field set, tokenizes and returns a document with
    all fields set
    """

    def __call__(self, doc: Document) -> Document:
        """
        Processes a document by tokenizing its text and setting relevant fields.

        Args:
            doc: A Document object with the text field set.

        Returns:
            A Document object with the following fields set:
            - text: The original text stripped of accents and converted to lowercase.
            - input_ids: List of token IDs created from the text.
            - token_mask: List of booleans indicating whether each token is a subword.
            - tokens: List of Token objects.
            - subword2word: Dictionary mapping subword indices to word indices.
            - dataset: A Dataset object created from the input IDs.
            - dataloader: A DataLoader object created from the dataset.
        """
        # get document's text and strip accent and lowercase
        doc.text = strip_accents_and_lowercase(doc.text)
        # create ids
        doc.input_ids = create_ids(doc.text)
        # create mask and tokens
        doc.token_mask, doc.tokens, doc.subword2word = create_mask_and_tokens(convert_to_tokens(doc.input_ids),
                                                                        remove_special_tokens(doc.input_ids))

        # create dataloader
        doc.dataset, doc.dataloader = create_dataset_and_dataloader(doc.input_ids)
        return doc



# The labels for the output of the dp model.
# A string label can be obtained by an output index
dp_labels = \
['obl',
 'obj',
 'dep',
 'mark',
 'case',
 'flat',
 'nummod',
 'obl:arg',
 'punct',
 'cop',
 'acl:relcl',
 'expl',
 'nsubj',
 'csubj:pass',
 'root',
 'advmod',
 'nsubj:pass',
 'ccomp',
 'conj',
 'amod',
 'xcomp',
 'aux',
 'appos',
 'csubj',
 'fixed',
 'nmod',
 'iobj',
 'parataxis',
 'orphan',
 'det',
 'advcl',
 'vocative',
 'compound',
 'cc',
 'discourse',
 'acl',
 'obl:agent']


import unittest

from transformers import AutoModel

from gr_nlp_toolkit.domain.document import Document
from gr_nlp_toolkit.processors.pos import POS
from gr_nlp_toolkit.processors.tokenizer import Tokenizer
from gr_nlp_toolkit.configs.pos_labels import pos_labels, pos_properties


class MyTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.bert_model = AutoModel.from_pretrained('nlpaueb/bert-base-greek-uncased-v1')


    def test_pos_with_one_example(self):
        tokenizer = Tokenizer()
        doc = tokenizer(Document('Ο ποιητής'))

        pos = POS()
        self.assertIsNotNone(pos._model)
        # self.assertIsNotNone(pos.system)
        doc = pos(doc)

        tokens = doc.tokens
        for token in tokens:
            self.assertIsNotNone(token.upos)
            self.assertTrue(token.upos in pos_labels['upos'])

            self.assertIsNotNone(token.feats)
            self.assertEqual(len(list(token.feats.keys())), len(pos_properties[token.upos]))

            for feat, value in token.feats.items():
                self.assertTrue(feat in pos_properties[token.upos])
                self.assertTrue(value in pos_labels[feat])

    def test_pos_with_one_example_with_subwords(self):
        tokenizer = Tokenizer()
        doc = tokenizer(Document('ενα ποιηματακι'))

        pos = POS(MyTestCase.bert_model)
        self.assertIsNotNone(pos._model)
        self.assertIsNotNone(pos.system)
        doc = pos(doc)

        tokens = doc.tokens
        for token in tokens:
            self.assertIsNotNone(token.upos)
            self.assertTrue(token.upos in pos_labels['upos'])

            self.assertIsNotNone(token.feats)
            self.assertEqual(len(list(token.feats.keys())), len(pos_properties[token.upos]))

            for feat, value in token.feats.items():
                self.assertTrue(feat in pos_properties[token.upos])
                self.assertTrue(value in pos_labels[feat])

    def test_pos_with_one_example_with_subwords(self):
        tokenizer = Tokenizer()
        doc = tokenizer(Document('ενα ποιηματακι'))

        pos = POS(MyTestCase.bert_model)
        self.assertIsNotNone(pos._model)
        self.assertIsNotNone(pos.system)
        doc = pos(doc)

        tokens = doc.tokens
        for token in tokens:
            self.assertIsNotNone(token.upos)
            self.assertTrue(token.upos in pos_labels['upos'])

            self.assertIsNotNone(token.feats)
            self.assertEqual(len(list(token.feats.keys())), len(pos_properties[token.upos]))

            for feat, value in token.feats.items():
                self.assertTrue(feat in pos_properties[token.upos])
                self.assertTrue(value in pos_labels[feat])

    def test_pos_with_one_example_with_subwords(self):
        tokenizer = Tokenizer()
        doc = tokenizer(Document('ενα ποιηματακι'))

        pos = POS()
        self.assertIsNotNone(pos._model)
        # self.assertIsNotNone(pos.system)
        doc = pos(doc)

        tokens = doc.tokens
        for token in tokens:
            self.assertIsNotNone(token.upos)
            self.assertTrue(token.upos in pos_labels['upos'])

            self.assertIsNotNone(token.feats)
            self.assertEqual(len(list(token.feats.keys())), len(pos_properties[token.upos]))

            for feat, value in token.feats.items():
                self.assertTrue(feat in pos_properties[token.upos])
                self.assertTrue(value in pos_labels[feat])


if __name__ == '__main__':
    unittest.main()



import math
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from gr_nlp_toolkit.configs.dictionary_tables import greeklish_to_greek_intonated

class LSTM_LangModel(nn.Module):
    """ 
    LSTM-based language model

    Attributes:
        hidden_size (int): The size of the hidden layer
        embed (nn.Embedding): The embedding layer
        lstm (nn.LSTM): The LSTM layer
        dense (nn.Linear): The dense layer
        dropout (nn.Dropout): The dropout layer
    """
    def __init__(self, input_size, embed_size, hidden_size, output_size):
        """
        Initializes the LSTM_LangModel with the specified parameters.

        Args:
            input_size (int): The size of the input layer
            embed_size (int): The size of the embedding layer
            hidden_size (int): The size of the hidden layer
            output_size (int): The size of the output layer
        """
        super(LSTM_LangModel, self).__init__()
        self.hidden_size = hidden_size

        self.embed = nn.Embedding(input_size, embed_size, padding_idx=0)
        self.lstm = nn.LSTM(embed_size, hidden_size, batch_first=True)
        self.dense = nn.Linear(hidden_size, output_size)
        self.dropout = nn.Dropout(0.5)

    def forward(self, x, h0=None, c0=None):
        """
        Forward pass of the LSTM_LangModel

        Args:
            x (Tensor): The input tensor
            h0 (Tensor): The initial hidden state
            c0 (Tensor): The initial cell state
        
        Returns:
            output (Tensor): The output tensor
            h (Tensor): The hidden state
            c (Tensor): The cell state
        
        """
        input_embedded = self.embed(x)
        if h0 is None and c0 is None:
            output_lstm, (h, c) = self.lstm(input_embedded)
        else:
            output_lstm, (h, c) = self.lstm(input_embedded, (h0, c0))
        output = self.dropout(output_lstm)
        output = self.dense(output)
        return output, h, c


class State():
    """
    Container class for the attributes of each candidate replacement.

    Attributes:
        translated (list): List of tokens already translated to the desired language.
        remaining (str): The remaining sentence to be translated.
        out (tensor): The last output of the LSTM for that particular state. Contains
                    logits that will become probabilities with the application of a softmax.
        hidden (tuple): Contains the hidden states of the candidate.
        score (float): The score given to the translation based on the language model.
    """

    def __init__(self, translated, remaining, out, hidden, score):

        """
        Initializes the State object with the specified parameters.

        Args:
            translated (list): List of tokens already translated to the desired language.
            remaining (str): The remaining sentence to be translated.
            out (tensor): The last output of the LSTM for that particular state. Contains
                    logits that will become probabilities with the application of a softmax.
            hidden (tuple): Contains the hidden states of the candidate.
            score (float): The score given to the translation based on the language model.
        """
        self.translated = translated
        self.remaining = remaining
        self.hidden = hidden
        self.output = out
        self.score = score

    def __eq__(self, other):
        """
        Equality operator, needed for eliminating duplicate states.
        """
        if isinstance(other, State):
            if (self.translated == other.translated and
                    self.remaining == other.remaining and
                    self.score == other.score):
                return True

        return False

class LanguageModel:
    """
    Language model for Greeklish to Greek conversion.

    Attributes:
        vectorizer (TextVectorizer): The vectorizer used to convert tokens to indices.
        model (LSTM_LangModel): The language model used for translation.
        device (str): The device to run the model on.
        softmax (nn.LogSoftmax): The log version of the softmax function.
    """

    def __init__(self, vectorizer, model, device='cpu'):
        """
        Initializes the LanguageModel with the specified parameters.

        Args:
            vectorizer: (TextVectorizer) The vectorizer used to convert tokens to indices.
            model: (LSTM_LangModel) The language model used for translation.
            device: (str) The device to run the model on.
        """
        self.vectorizer = vectorizer
        self.model = model
        self.mode = vectorizer.mode
        self.device = torch.device(device)
        self.model.to(self.device)

        # Use the log version of Softmax to sum scores instead of multiplying them and avoid decay.
        self.softmax = nn.LogSoftmax(dim=1)

    def load_model(self, path):
        """
        Load a pre-trained model as a state dictionary.

        Args:
            path (str): The path to the pre-trained model.
        """
        self.model.load_state_dict(torch.load(path, weights_only=True))

    def translate(self, sentences, beams):
        """
        Takes a list of sentences and translates them.

        Args:
            sentences: (list) Sentences you want to translate
            beams: (int) The number of parameters

        Returns:
         translated_sentences: (list) Translated sentences

        """
        # Don't forget to put the model in eval mode.
        self.model.eval()
        with torch.no_grad():
            translated_sentences = []
            for sentence in sentences:
                translated = []
                remaining = sentence
                # We start with the first state, with a score of 1
                # The format of a state is: (translated_sent, remaining_sent, (h0, c0), score)
                # --------------------------------------------------------------------------------
                # First, we need to "prep" the char model. This is done by feeding the network with the
                # <s> token and saving the output hidden states for the initial State() object.
                start_input = self.vectorizer.input_tensor("<s>")
                out, h_n, c_n = self.model(start_input.to(self.device), None, None)
                # The score of the initial state in 0, because we use LogSoftmax instead of regular Softmax.
                initial_state = State(translated, remaining, out, (h_n, c_n), 0)
                states = [initial_state]
                for i in range(len(sentence)):
                    candidates = []
                    # Look through the current states.
                    for state in states:
                        # Produce the next-char-candidates from each state, along with their probabilities.
                        new_states = self.get_candidates(state)
                        candidates += new_states

                    # Remove any duplicate candidates
                    #print([candidate.translated for candidate in candidates])
                    candidates_set = []
                    [candidates_set.append(cand) for cand in candidates if cand not in candidates_set]
                    candidates = candidates_set
                    # Get the best states, according to the number of beams in the search.
                    best_candidates = []
                    for j in range(beams):
                        if candidates:
                            # Probabilities of each candidate new state.
                            probabilities = [cand.score for cand in candidates]
                            # Get the best candidate and remove from the list
                            best_cand = candidates.pop(probabilities.index(max(probabilities)))
                            # Add that candidate to the list of best candidates
                            best_candidates.append(best_cand)
                    # Make the list of the best candidates the new states.
                    states = best_candidates

                # Once the sentence is over, the state with the highest probability is the best translation.
                probs = [state.score for state in states]
                # Extract the sentence from the state
                sent = states[probs.index(max(probs))]
                # Convert the list of translated tokens to a sentence.
                translation = ""
                for i in sent.translated:
                    translation += i

                translated_sentences.append(translation)
            return translated_sentences

    def get_candidates(self, state):
        """
        Get the next candidates for the translation.

        Args:
            state (State): The current state of the translation.
        
        Returns:
            candidates (list): A list of the next candidates for the translation.

        """


        # If the state is already a final state (no remaining text to translate),
        # it returns only itself as a candidate.
        if not state.remaining:
            return [state]

        # If it is not a final state, generate the next candidate states
        candidates = []

        # Look at both of the first two characters in the translated sentence, as some greek
        # characters may be represented with 2 characters in Greeklish.
        for length in [1, 2]:
            if len(state.remaining) >= length:
                # Fetch the valid replacements from the dictionary.
                if length == 2:
                    token = state.remaining[0] + state.remaining[1]
                    replacements = greeklish_to_greek_intonated.get(token, [])
                else:
                    # If the look-up is a miss (e.g. the token is a space, a number or punctuation),
                    # return the token itself.
                    token = state.remaining[0]
                    replacements = greeklish_to_greek_intonated.get(token, [token])

                # For each candidate replacement, get the probability from the LM
                for item in replacements:
                    h_n, c_n = state.hidden[0], state.hidden[1]
                    out = state.output
                    score = state.score
                    for token in item:
                        # Apply softmax to the model's output and get the prob based on the index from vocab
                        probs = self.softmax(out)
                        idx = self.vectorizer.vocab.get(token, len(self.vectorizer.vocab))
                        # Update score.
                        score = score + probs[0][idx].item()
                        # Feed the token to the model to get the next output and hidden states
                        input = self.vectorizer.input_tensor(token)
                        out, h_n, c_n = self.model(input.to(self.device), h_n, c_n)

                    translated_tokens = [token for token in item]

                    new_candidate = State(state.translated+translated_tokens,
                                          state.remaining[length:],
                                          out, (h_n, c_n), score)
                    candidates.append(new_candidate)

        return candidates


from torch.nn import LeakyReLU

from gr_nlp_toolkit.models.util import create_mask_from_length

from torch import nn
import torch


class DPModel(nn.Module):
    """
    Dependency Parsing model.

    Attributes:
        numrels (int): Number of dependency relation labels.
        _bert_model (nn.Module): The BERT model.
        _dp (nn.Dropout): Dropout layer.
        arc_head (nn.Linear): Linear layer for arc head representation.
        arc_dep (nn.Linear): Linear layer for arc dependent representation.
        rel_head (nn.Linear): Linear layer for relation head representation.
        rel_dep (nn.Linear): Linear layer for relation dependent representation.
        arc_bias (nn.Parameter): Bias parameter for arc representation.
        rel_bias (nn.Parameter): Bias parameter for relation representation.
        u_rel (nn.Parameter): Parameter for relation representation.
        w_arc (nn.Parameter): Parameter for arc representation.
        w_rel_head (nn.Parameter): Parameter for relation head representation.
        w_rel_dep (nn.Parameter): Parameter for relation dependent representation.
        deprel_linear_2 (nn.Linear): Linear layer for dependency relation labels.
        relu (LeakyReLU): LeakyReLU activation function.
    """

    def __init__(self, bert_model, deprel_i2l, dp):
        """
        Initialize the DPModel.

        Args:
            bert_model (nn.Module): The BERT model.
            deprel_i2l (list): List of dependency relation labels.
            dp (float): The dropout probability.

        """
        super(DPModel, self).__init__()
        self.numrels = len(deprel_i2l)
        self._bert_model = bert_model
        self._dp = nn.Dropout(dp)

        self.arc_head = nn.Linear(768, 768)
        self.arc_dep = nn.Linear(768, 768)

        self.rel_head = nn.Linear(768, 768)
        self.rel_dep = nn.Linear(768, 768)

        self.arc_bias = nn.Parameter(torch.zeros(1, 768, 1))
        self.rel_bias = nn.Parameter(torch.zeros(1, 1, 1, self.numrels))

        self.u_rel = nn.Parameter(torch.zeros(1, 768, self.numrels * 768))

        self.w_arc = nn.Parameter(torch.zeros(1, 768, 768))
        self.w_rel_head = nn.Parameter(torch.zeros(1, 1, 768, self.numrels))
        self.w_rel_dep = nn.Parameter(torch.zeros(1, 1, 768, self.numrels))

        self.deprel_linear_2 = nn.Linear(768, len(deprel_i2l) * 768)

        self.relu = LeakyReLU(1)


    def forward(self, text, text_len):
        """
        Forward pass of the DPModel.

        Args:
            text (Tensor): Input text.
            text_len (Tensor): Length of the input text.

        Returns:
            output (dict): Dictionary containing the output of the model.

        """
        output = {}

        attention_mask = create_mask_from_length(text_len, text.shape[1])
        bert = self._bert_model(text, attention_mask=attention_mask)

        # output size bs , mseq , 768
        bert_output = self._dp(bert[0])
        bs = bert_output.shape[0]
        mseq = bert_output.shape[1]

        # Specialized vector representations
        arc_head = self.relu(self.arc_head(bert_output))  # bs,mseq,768
        arc_dep = self.relu(self.arc_dep(bert_output))  # bs,mseq,768
        rel_head = self.relu(self.rel_head(bert_output))  # bs,mseq,768
        rel_dep = self.relu(self.rel_dep(bert_output))  # bs,mseq,768

        # bs,mseq,768 @ bs,768,mseq + bs,mseq,768 @ 1,768,1
        output_linear_head = arc_head @ (arc_dep @ self.w_arc).transpose(1, 2) + arc_head @ self.arc_bias
        # arcdep * self w.arc  = (bs,mseq,768) * (1,768,768) = (bs, mseq , 768)

        # bs,mseq, 768 * 1,768,768 *numrel = bs,mseq,numrel,768,`
        label_biaffine = rel_dep @ self.u_rel # bs,mseq,768 * numrel
        label_biaffine = label_biaffine.reshape(bs,mseq,self.numrels,768)
        label_biaffine = label_biaffine @ rel_head.transpose(1,2).unsqueeze(1) # bs,mseq,numrel,mseq
        label_biaffine = label_biaffine.transpose(2,3)

        label_head_affine = (rel_head.unsqueeze(2) @ self.w_rel_head)
        label_dep_affine = (rel_dep.unsqueeze(2) @ self.w_rel_dep)
        label_bias = self.rel_bias

        output_linear_rel = label_biaffine + label_head_affine + label_dep_affine + label_bias
        #                   (bs,mseq,1,768) @ (1 , 1 , 768 ,numrels) + ( 1, 1 , 1, numrels)
        #                    (bs,mseq,1,numrels)

        output['heads'] = output_linear_head
        output['deprels'] = output_linear_rel.reshape(bs, mseq, mseq, self.numrels)

        selected_arcs = output_linear_head.argmax(-1)  # bs,mseq (indexes in [0,mseq) )
        selected_arcs = selected_arcs.unsqueeze(-1).repeat(1, 1, mseq)  # bs,mseq,mseq
        selected_arcs = selected_arcs.unsqueeze(-1).repeat(1, 1, 1, self.numrels)  # bs,mseq,mseq, numrels

        deprels_output = torch.gather(output_linear_rel, dim=2, index=selected_arcs)  # bs,mseq,mseq,numrels
        # dim 2 is redundant so must be deleted ( there is only one head for every token)
        deprels_output = deprels_output.narrow(2, 0, 1)  # bs,mseq,1,numrels
        deprels_output = deprels_output.squeeze(2)  # bs , mseq,numrels
        output['gathered_deprels'] = deprels_output

        return output


# This is a dict where each entry contains an label for a morphological feature,
# or the label for the UPOS tag if the key is 'upos'
pos_properties = {'ADJ': ['Degree', 'Number', 'Gender', 'Case'],
 'ADP': ['Number', 'Gender', 'Case'],
 'ADV': ['Degree', 'Abbr'],
 'AUX': ['Mood',
         'Aspect',
         'Tense',
         'Number',
         'Person',
         'VerbForm',
         'Voice'],
 'CCONJ': [],
  'DET': ['Number', 'Gender', 'PronType', 'Definite', 'Case'],
  'NOUN': ['Number', 'Gender', 'Abbr', 'Case'],
  'NUM': ['NumType', 'Number', 'Gender', 'Case'],
  'PART': [],
  'PRON': ['Number', 'Gender', 'Person', 'Poss', 'PronType', 'Case'],
  'PROPN': ['Number', 'Gender', 'Case'],
  'PUNCT': [],
  'SCONJ': [],
  'SYM': [],
  'VERB': ['Mood',
          'Aspect',
          'Tense',
          'Number',
          'Gender',
          'Person',
          'VerbForm',
          'Voice',
          'Case'],
  'X': ['Foreign'],
  '_': []}

# The labels for the named entity output of the ner model.
# A string label can be obtained by an output index
pos_labels = {'Abbr': ['_', 'Yes'],
 'Aspect': ['Perf', '_', 'Imp'],
 'Case': ['Dat', '_', 'Acc', 'Gen', 'Nom', 'Voc'],
 'Definite': ['Ind', 'Def', '_'],
 'Degree': ['Cmp', 'Sup', '_'],
 'Foreign': ['_', 'Yes'],
 'Gender': ['Fem', 'Masc', '_', 'Neut'],
 'Mood': ['Ind', '_', 'Imp'],
 'NumType': ['Mult', 'Card', '_', 'Ord', 'Sets'],
 'Number': ['Plur', '_', 'Sing'],
 'Person': ['3', '1', '_', '2'],
 'Poss': ['_', 'Yes'],
 'PronType': ['Ind', 'Art', '_', 'Rel', 'Dem', 'Prs', 'Ind,Rel', 'Int'],
 'Tense': ['Pres', 'Past', '_'],
 'VerbForm': ['Part', 'Conv', '_', 'Inf', 'Fin'],
 'Voice': ['Pass', 'Act', '_'],
 'upos': ['X',
          'PROPN',
          'PRON',
          'ADJ',
          'AUX',
          'PART',
          'ADV',
          '_',
          'DET',
          'SYM',
          'NUM',
          'CCONJ',
          'PUNCT',
          'NOUN',
          'SCONJ',
          'ADP',
          'VERB']}



import torch
from torch import nn

from transformers import AutoModel

from gr_nlp_toolkit.domain.document import Document
from gr_nlp_toolkit.processors.abstract_processor import AbstractProcessor
from gr_nlp_toolkit.configs.pos_labels import pos_labels, pos_properties


from gr_nlp_toolkit.models.pos_model import POSModel


class POS(AbstractProcessor):
    """
    Part-Of-Speech (POS) processor class.

    This class performs POS tagging using a pre-trained BERT model. It initializes the model,
    loads the necessary components, and provides functionality to process documents
    and assign POS tags and features to tokens.

     Attributes:
        properties_POS (dict): Dictionary containing properties for POS tags.
        feat_to_I2L (dict): Dictionary mapping feature names to label lists.
        feat_to_size (dict): Dictionary mapping feature names to the size of their label lists.
        _model (POSModel): The POS model based on BERT.
        softmax (nn.Softmax): Softmax function for output normalization.
        device (torch.device): Device on which the model is loaded.
    """

    def __init__(self, model_path=None, device='cpu'):
        """
        Initializes the POS class with the specified parameters.

        Args:
            model_path (str, optional): Path to the pre-trained model. Defaults to None.
            device (str, optional): Device to load the model on ('cpu' or 'cuda'). Defaults to 'cpu'.
        """

        self.properties_POS = pos_properties
        self.feat_to_I2L = pos_labels
        self.feat_to_size = {k: len(v) for k, v in self.feat_to_I2L.items()}

        # model init
        bert_model = AutoModel.from_pretrained('nlpaueb/bert-base-greek-uncased-v1')
        self._model = POSModel(bert_model, self.feat_to_size, 0)
        self.softmax = nn.Softmax(dim=-1)
        self.device = torch.device(device)
        self._model.to(self.device)
        self._model.eval()

        # load the pretrained model
        if model_path is not None:
            self._model.load_state_dict(torch.load(model_path, map_location=self.device, weights_only=True), strict=False)

    def __call__(self, doc: Document) -> Document:
        """
        Processes a document to perform Part-Of-Speech tagging and assign features.

        Args:
            doc (Document): The document to process.

        Returns:
            Document: The document with POS tags and features assigned to the tokens.
        """

        predictions = {}

        input_ids, text_len = next(iter(doc.dataloader))['input']

        for feat in self.feat_to_I2L.keys():
            output = self._model(input_ids.to(self.device), text_len.to(self.device))
            output = self.softmax(output[feat])

            
            predictions[feat] = torch.argmax(output[0], axis=-1).detach().cpu().numpy()

        # set upos
        upos_predictions = predictions['upos']
        i = 0
        for mask, pred in zip(doc.token_mask, upos_predictions[1: len(upos_predictions) - 1]):
            if mask:
                token = doc.tokens[i]
                token.upos = self.feat_to_I2L['upos'][pred]
                # Advance to the next word (not subtoken)
                i+=1

        # set features
        for feat in self.feat_to_I2L.keys():
            if feat != 'upos':
                current_predictions = predictions[feat]
                i = 0
                for mask, pred in zip(doc.token_mask, current_predictions[1: len(current_predictions) - 1]):
                    if mask:
                        token = doc.tokens[i]
                        if feat in self.properties_POS[token.upos]:
                            token.feats[feat] = self.feat_to_I2L[feat][pred]
                        # Advance to the next word (not subtoken)
                        i += 1

        return doc



import unittest

from gr_nlp_toolkit.domain.token import Token


class MyTestCase(unittest.TestCase):
    def test_new_token_object(self):
        token = Token(['α'])
        self.assertEqual(['α'], token.subwords)


if __name__ == '__main__':
    unittest.main()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gr-nlp-toolkit",
    version="0.1.1",
    author="nlpaueb",
    author_email="p3170148@aueb.gr, p3170039@aueb.gr, spirosbarbakos@gmail.com,  eleftheriosloukas@aueb.gr, ipavlopoulos@aueb.gr",
    description="A Transformer-based Natural Language Processing Pipeline for Greek",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nlpaueb/gr-nlp-toolkit",
    project_urls={
        #"Bug Tracker": "https://github.com/pypa/sampleproject/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: Linguistic",
        "Natural Language :: Greek"
    ],
    packages=setuptools.find_packages(where=".", exclude="./test_gr_nlp_toolkit"),
    python_requires=">=3.9",
    install_requires=[
         'torch>=2.1.2',
         'transformers>=4.11.1',
         'huggingface_hub'
    ]
)



import unittest

from transformers import AutoModel

from gr_nlp_toolkit.domain.document import Document
from gr_nlp_toolkit.processors.dp import DP
from gr_nlp_toolkit.processors.tokenizer import Tokenizer

from gr_nlp_toolkit.configs.dp_labels import dp_labels


class MyTestCase(unittest.TestCase):

    def test_dp_with_one_example(self):
        tokenizer = Tokenizer()
        doc = tokenizer(Document('Ο ποιητής'))

        dp = DP()
        self.assertIsNotNone(dp._model)

        doc = dp(doc)

        tokens = doc.tokens
        for token in tokens:
            self.assertIsNotNone(token.head)
            self.assertIsNotNone(token.deprel)
            self.assertTrue(token.head in range(0, len(tokens)))
            self.assertTrue(token.deprel in dp_labels)

    def test_dp_with_one_example_with_subwords(self):
        tokenizer = Tokenizer()
        doc = tokenizer(Document('ενα ποιηματακι'))

        # bert model init
        dp = DP()

        self.assertIsNotNone(dp._model)
        doc = dp(doc)

        tokens = doc.tokens
        for token in tokens:
            self.assertIsNotNone(token.head)
            self.assertIsNotNone(token.deprel)
            self.assertTrue(token.head in range(0, len(tokens)))
            self.assertTrue(token.deprel in dp_labels)


if __name__ == '__main__':
    unittest.main()


import unittest

from gr_nlp_toolkit.configs.dp_labels import dp_labels
from gr_nlp_toolkit.configs.ner_labels import ner_labels
from gr_nlp_toolkit.configs.pos_labels import pos_labels, pos_properties
from gr_nlp_toolkit.pipeline.pipeline import Pipeline



class TestPipeline(unittest.TestCase):
    # The unit tests test the transformer-based g2g processor. If you want to test the LSTM-based processor, you can change 'g2g' to 'g2g_lite'
    def test_using_all_processors(self):
        nlp = Pipeline('dp,pos,ner,g2g') 

        sentences = ["Η Ιταλία κέρδισε την Αγγλία στον τελικό του Euro το 2021",
                     "Το ποιηματάκι το έγραψε ο διάσημος ποιητής, Νίκος Νικολαϊδης",
                     "Uparxoun autoi pou kerdizoun apo mia katastash kai autoi pou xanoun"]
        for sent in sentences:
            doc = nlp(sent)

            for token in doc.tokens:
                print(token.text, token.ner, token.upos, token.feats, token.head, token.deprel)
                self.assertIsNotNone(token.ner)
                self.assertTrue(token.ner in ner_labels)
                self.assertIsNotNone(token.head)
                self.assertIsNotNone(token.deprel)
                # We have to add plus one, because the cls token is removed
                self.assertTrue(token.head in range(0, len(doc.tokens) + 1))
                self.assertTrue(token.deprel in dp_labels)
                self.assertIsNotNone(token.upos)
                self.assertTrue(token.upos in pos_labels['upos'])

                self.assertIsNotNone(token.feats)
                self.assertEqual(len(list(token.feats.keys())), len(pos_properties[token.upos]))

                for feat, value in token.feats.items():
                    self.assertTrue(feat in pos_properties[token.upos])
                    self.assertTrue(value in pos_labels[feat])
                    print(token.text, token.ner, token.upos, token.feats, token.head, token.deprel)
                    self.assertIsNotNone(token.ner)
                    self.assertTrue(token.ner in ner_labels)
                    self.assertIsNotNone(token.head)
                    self.assertIsNotNone(token.deprel)
                    # We have to add plus one, because the cls token is removed
                    self.assertTrue(token.head in range(0, len(doc.tokens) + 1))
                    self.assertTrue(token.deprel in dp_labels)
                    self.assertIsNotNone(token.upos)
                    self.assertTrue(token.upos in pos_labels['upos'])

    def test_annotations_are_same_with_multiple_configurations(self):
        nlp = Pipeline('dp,pos,ner,g2g')
        doc = nlp("Uparxoun autoi pou kerdizoun apo mia katastash kai autoi pou xanoun")

        deprels_preds = []
        upos_preds = []
        ner_preds = []
        for token in doc.tokens:
            deprels_preds.append(token.deprel)
            upos_preds.append(token.upos)
            ner_preds.append(token.ner)

        nlp = Pipeline('dp,g2g')
        doc = nlp("Uparxoun autoi pou kerdizoun apo mia katastash kai autoi pou xanoun")
        new_deprels_preds = []

        for token in doc.tokens:
            new_deprels_preds.append(token.deprel)

        nlp = Pipeline('pos,g2g')
        doc = nlp("Uparxoun autoi pou kerdizoun apo mia katastash kai autoi pou xanoun")
        new_upos_preds =[]

        for token in doc.tokens:
            new_upos_preds.append(token.upos)

        nlp = Pipeline('ner,g2g')
        doc = nlp("Uparxoun autoi pou kerdizoun apo mia katastash kai autoi pou xanoun")
        new_ner_preds =[]
        for token in doc.tokens:
            new_ner_preds.append(token.ner)

        self.assertEqual(new_deprels_preds, deprels_preds)
        self.assertEqual(new_upos_preds, upos_preds)
        self.assertEqual(new_ner_preds, ner_preds)



    def test_using_only_one_processor(self):
        nlp = Pipeline('ner')
        doc = nlp("Η Ιταλία κέρδισε την Αγγλία στον τελικό του Euro το 2021")

        for token in doc.tokens:
            self.assertIsNotNone(token.ner)
            self.assertTrue(token.ner in ner_labels)
            self.assertIsNone(token.head)
            self.assertIsNone(token.deprel)
            self.assertFalse(token.head in range(0, len(doc.tokens)))
            self.assertFalse(token.deprel in dp_labels)
            self.assertIsNone(token.upos)
            self.assertFalse(token.upos in pos_labels['upos'])

            for feat, value in token.feats.items():
                self.assertFalse(feat in pos_properties[token.upos])
                self.assertFalse(value in pos_labels[feat])


if __name__ == '__main__':
    unittest.main()



import unittest

from transformers import AutoModel

from gr_nlp_toolkit.domain.document import Document
from gr_nlp_toolkit.processors.ner import NER
from gr_nlp_toolkit.processors.tokenizer import Tokenizer

from gr_nlp_toolkit.configs.ner_labels import ner_labels


class MyTestCase(unittest.TestCase):

    def test_ner_with_one_example(self):
        tokenizer = Tokenizer()
        doc = tokenizer(Document('Ο ποιητής'))

        ner = NER(entities=18)

        self.assertEqual(69, ner.output_size)
        self.assertIsNotNone(ner._model)
        doc = ner(doc)

        tokens = doc.tokens
        for token in tokens:
            self.assertIsNotNone(token.ner)
            self.assertTrue(token.ner in ner_labels)

    def test_ner_with_one_example_with_subwords(self):
        tokenizer = Tokenizer()
        doc = tokenizer(Document('ενα ποιηματακι'))

        ner = NER()
        self.assertIsNotNone(ner._model)
        doc = ner(doc)

        tokens = doc.tokens
        for token in tokens:
            self.assertIsNotNone(token.ner)
            self.assertTrue(token.ner in ner_labels)

    def test_ner_with_value_exception(self):
        with self.assertRaises(ValueError):
            NER(entities=2)


if __name__ == '__main__':
    unittest.main()



import unittest

from gr_nlp_toolkit.processors.g2g import G2G
from gr_nlp_toolkit.processors.g2g import detect_language
from gr_nlp_toolkit.domain.document import Document

class MyTestCase(unittest.TestCase):

    def test_g2g_lstm(self):

        g2g = G2G(mode="LSTM",model_path="gr_nlp_toolkit/RBNLM_weights/LSTM_LM_50000_char_120_32_512.pt", tokenizer_path="gr_nlp_toolkit/RBNLM_weights/RBNLMtextVectorizer.pkl")
        self.assertIsNotNone(g2g.model)
        self.assertIsNotNone(g2g.text_vectorizer)
        self.assertIsNotNone(g2g.LM)


        doc = Document("o volos kai h larisa einai poleis ths thessalias")
        doc = g2g(doc)
        self.assertEqual(detect_language(doc.text), 'greek')
        self.assertEqual(doc.text.split()[5], "είναι")

    
    def test_g2g_transformer(self):
            
        g2g = G2G(mode="transformer", model_path="AUEB-NLP/ByT5_g2g")
        self.assertIsNotNone(g2g.model)

        doc = Document('"o volos kai h larisa einai poleis ths thessalias"')
        doc = g2g(doc)
        self.assertEqual(detect_language(doc.text), 'greek')
        self.assertEqual(doc.text.split()[5], "είναι")

        
if __name__ == '__main__':
    unittest.main()



import torch


def create_mask_from_length(length_tensor, mask_size):

    """
    Creates a binary mask based on length.

    Args:
        length_tensor (torch.Tensor): ND Tensor containing the lengths.
        mask_size (int): Integer specifying the mask size. Usually the largest length in the batch
    
    Return:
        torch.Tensor (N+1)D Int Tensor (..., mask_size) containing the binary mask.
    """

    mask = torch.arange(0, mask_size, dtype=torch.int, device=length_tensor.device)
    
    mask = mask.int().view([1] * (len(length_tensor.shape)) + [-1])

    return mask < length_tensor.int().unsqueeze(-1)

